;;; Programa: Package con agentes jugadores humanos de PlanTron
;;;
;;; Asignatura: Introducción a la Inteligencia Artificial
;;;             ETSII - ULL
;;; Práctica: (sistemas de producción)
;;; Profesor: Patricio García (patricio AT etsii.ull.es)
;;; Fecha:    12/04/2006
;;; Comentarios:
;;;
;;;           El agente definido utiliza el entorno de juego
;;;           PlanTron, definido principalmente por los packages
;;;           pt-base y envio.
;;;
;;;           Las funciones accesibles más importantes son:
;;;
;;;             AGENTE-HUMANO
;;;             AGENTE-HUMANO-INI
;;;             AGENTE-HUMANO-FIN
;;;
;;;           Ver la documentación existente para cada una de ellas.
;;;
;;;           Para su utilización es aconsejable usar el cargador
;;;           del entorno de juego PlanTron: "pt-loader.lisp".
;;;
;;;
;;; Dependencias:
;;;
;;;           Hace uso del paquete que contiene la interface (ENVIO)
;;;           con el control del juego (PT-BASE).
;;;
;;;           También resulta prácticamente inútil si no se utiliza
;;;           junto a un agente visualizador, ver pt-vis-*.lisp.
;;;
;;;
;;;           Debido a la necesidad de controlar el teclado
;;;           por parte de los jugadores humanos este entorno
;;;           tiene limitaciones en gcl, de momento (ver bug)
;;;
;;;           También tiene problemas con clisp en versiones < 2.33
;;;           debido a bug en la función read-char-no-hang (hace hang),
;;;           ver: http://groups.google.com/group/comp.lang.lisp/browse_frm/thread/6bfac6923a4126b1/5575c9088e5ca9c3?tvc=1
;;;
;;;
;;; Modificaciones:
;;;
;;;           - Ver 0.1 (13/04/2006): Primera versión. Extraida
;;;             de tron.0.1.lsp. Vease tron-base.lsp y tron-comm*.lsp
;;;
;;;           - Ver 0.2 (16/04/2006): Generada versión para clisp,
;;;             ver echo-hagn-off-clisp.lsp. Arreglado problemas de
;;;             comido de teclas en juego de dos agentes humanos (ver
;;;             unread-char-no-hang-no-echo). Arreglado problema de
;;;             reactivación del echo en CMU lisp.
;;;
;;;           - Ver 0.2.1 (21/04/2006): Eliminados warnings de CMUCL
;;;             (ver "(declare (ignore ..."). Añadida documentación.
;;;
;;;           - Ver 0.3 (29/04/2006): Empaquetado, puesto interface
;;;             de definición de agentes (const agentes). Ofuscado
;;;             código de agente-humano.
;;;
;;;           - Ver 0.4 (03/05/2006): Eliminado bug que impedía jugar
;;;             un agente humano solo.
;;;
;;;           - Ver 0.5 (03/05/2006):  Cambio en agente-humano-fin
;;;             debido a cambio en tron-comm de ver "fun-0.2".
;;;
;;;           - Ver 0.5.1 (05/05/2006): Inclusión de warning si se
;;;             carga con GCL. 
;;;
;;;           - Ver 0.5.2 (12/04/2007): Generada versión para sbcl,
;;;             ver echo-hagn-off-sbcl.lsp.
;;;
;;;           - Ver 0.6 (21/4/2007): Cambio de nombre de 
;;;             tron-agentes-humanos a pt-agentes-humanos. Cambio 
;;;             extensión de "lsp" a "lisp". Cambio también nombre de
;;;             variables, funciones, etc de "tron" a "pt".
;;;
;;;           - Ver 0.7 (1/05/2007): Ajustes para versión con nueva
;;;             interface de comunicaciones en package "ENVIO"
;;;
;;;           - Ver 0.7.1 (25/05/2007): Modificada llamada a refresco
;;;             mediante uso del servicio :refresco
;;;
;;;           - Ver 0.8 (09/04/2009): Definicíon usando defagent
;;;
;;;
;;; Por hacer:
;;;           - Corrección de Bugs
;;;             ------------------
;;;           - Problema con agente humano gcl que necesita NL
;;;             despúes de pulsar las teclas de movimiento
;;;
;;;           - Mejoras en funcionalidades
;;;             --------------------------
;;;           - ¿Cambiar teclas jugado humano 1?
;;;           - Prepararlo para otras variantes de lisp (sbcl,¿Windows?)
;;;           - Agregar agente-humano-asistido, sólo acepte movimientos
;;;             válidos
;;;           - Eliminar notas y warnings de versión se sbcl
;;;
;;;           - Mejoras en estilo de programación
;;;             ---------------------------------
;;;           - Que acepte Ctr-L para refrescar pantalla (#\Page en sbcl)
;;;           - Cambiar cálculos de lit por #+cmu ...
;;;           - Hacer pasar fun-refresco por envio con :refresco
;;;
;;;


;;
;; Definición del package
;; ----------------------

(defpackage "PT-AGENTES-HUMANOS"
  (:nicknames "PT-A-H" "PT-A-H")
  (:use "COMMON-LISP")
  (:export "AGENTES" "VERSION" "AGENTE-HUMANO" "AGENTE-HUMANO-INI" "AGENTE-HUMANO-FIN"))

(in-package "PT-AGENTES-HUMANOS")


;;
;; Constante versión
;; -----------------

(defconstant version "0.8"
  "Versión del package PT-AGENTES-HUMANOS.")


;;
;; Variables globales
;; ------------------

(defvar echo-hang t
  "Variable que indica estado echo-hang de la salida/entrada.")


;;
;; Funciones específicas según implementación lisp
;; -----------------------------------------------

(let ((lit (lisp-implementation-type)))
  (cond
   ((string-equal lit "GNU Common Lisp (GCL)")
    ; GNU gcl
    ; -------
    (envio:my-warning "En esta versión ~A existe un problema con la lectura de pulsaciones de teclas para el agente humano, no son leidas hasta que no se pulse la tecla enter." lit)
    (load (concatenate 'string
		     (directory-namestring
		      (eval (or (find-symbol "*LOAD-TRUENAME*")
				(find-symbol "*LOAD-PATHNAME*")
				(find-symbol "*LOAD-PATHNAME*" "SYSTEM")
			       "")))
		     "echo-hang-off-gcl")))

  ((string-equal lit "CMU Common Lisp")
    ; CMUCL
    ; -----
   (load (concatenate 'string
		     (directory-namestring
		      (eval (or (find-symbol "*LOAD-TRUENAME*")
				(find-symbol "*LOAD-PATHNAME*")
				(find-symbol "*LOAD-PATHNAME*" "SYSTEM")
			       "")))
		     "echo-hang-off-cmu")))

   ((string-equal lit "CLISP")
     ; GNU clisp
     ; ---------
    (load (concatenate 'string
		     (directory-namestring
		      (eval (or (find-symbol "*LOAD-TRUENAME*")
				(find-symbol "*LOAD-PATHNAME*")
				(find-symbol "*LOAD-PATHNAME*" "SYSTEM")
			       "")))
		     "echo-hang-off-clisp")))

   ((string-equal lit "SBCL")
    ; SBCL
    ; ----
   (load (concatenate 'string
		     (directory-namestring
		      (eval (or (find-symbol "*LOAD-TRUENAME*")
				(find-symbol "*LOAD-PATHNAME*")
				(find-symbol "*LOAD-PATHNAME*" "SYSTEM")
			       "")))
		     "echo-hang-off-sbcl")))

   (t
     ; Desconocida
     ; -----------
    (format *error-output* "Implementación ~S no soportada" lit)

    (defun echo-hang-off ()
      (setf echo-hang nil))

    (defun restore-echo-hang ()
      (setf echo-hang t))

    (defun read-char-no-hang-no-echo (&optional (stream *standard-input*))
      (read-char-no-hang stream))

    (defun unread-char-no-hang-no-echo (char &optional (stream *standard-input*))
      (unread-char char stream)))))


;;
;; Agente humano básico
;; --------------------

(defconstant tecla-mov-1
  ; Jugador 1: mov. vi
  '((#\k 0) (#\K 0) (#\l 1) (#\L 1) (#\j 2) (#\J 2) (#\h 3) (#\H 3))
  "Lista asociativa con las teclas y movimientos del jugador 1.")

(defconstant tecla-mov-def
  ; Jugador 2 o por defecto: mov. cursores
  '((#\8 0) (#\A 0) (#\6 1) (#\C 1) (#\2 2) (#\B 2) (#\4 3) (#\D 3))
  "Lista asociativa con las teclas y movimientos del jugador por defecto o jugador 2 si juegan 2 humanos.")

(defvar tecla-mov-jugs nil
  "Lista asociativa con jugadores y lista asociativa de las teclas y movimientos del cada uno.")

(defun refresco-tablero ()
  "Llamada función para refrescar el tablero."
  (envio:broadcast :refresco))

(defun agente-humano-ini (jug tipo-juego)
  "Función que inicializa el agente-humano."
  (declare (ignore tipo-juego))
  (if echo-hang
      (progn
	(setf tecla-mov-jugs (list (cons jug tecla-mov-def)))
	; Ponemos echo a off, evitar bloquear lecturas y
        ; leer inmediatamente
	(echo-hang-off))
    ; Para dos jugadores humanos (no comer carácteres de oponente)
    (setf tecla-mov-jugs (list (cons 1 tecla-mov-1) (cons 2 tecla-mov-def))))
  (clear-input)
  t)

(defun agente-humano-fin (jug resultado-juego)
  "Función que cierra la ejecución del agente-humano."
  (declare (ignore jug resultado-juego))
  (if (not echo-hang)
      (progn
	; Restauramos estado tty y limpiamos buffer
	(restore-echo-hang)
	(sleep 2)
	(clear-input)))
  t)

(defun agente-humano (jugador)
  "Agente humano que lee los movimientos del jugador Pt JUGADOR (1 ó 2) desde teclado. Solo utiliza las teclas del cursor (8, 6, 2, 4), cuando se llama con dos agentes humanos utiliza para el jugador 1 (izquierdo) las teclas de movimientos del vi(1) (k, l, j, h), y para el jugador 2 (derecho) las teclas del cursor. La barra espaciador refresca la pantalla, siempre que se tenga cargado el agente de visualización del package pt-vis."
  (let* ((car (read-char-no-hang-no-echo))
	 (mov (first (rest (assoc car (rest (assoc jugador tecla-mov-jugs)))))))
    (cond (mov mov)
	  ; Refresco pantalla
	  ((eq car #\Space)
	   (refresco-tablero) nil)
	  ; Devolvemos car si no era para nosotros, far play
	  ((assoc car (rest (assoc (- 3 jugador) tecla-mov-jugs)))
	   (unread-char-no-hang-no-echo car) nil))))


;;
;;Definición de agente presentes en este package
;; ---------------------------------------------

; Agente humano, efectua los movimientos indicados por teclado.

(pt:defagente humano
  "Agente humano que lee los movimientos del jugador Pt JUGADOR (1 ó 2) desde teclado. Solo utiliza las teclas del cursor (8, 6, 2, 4), cuando se llama con dos agentes humanos utiliza para el jugador 1 (izquierdo) las teclas de movimientos del vi(1) (k, l, j, h), y para el jugador 2 (derecho) las teclas del cursor. La barra espaciador refresca la pantalla, siempre que se tenga cargado el agente de visualización del package pt-vis"
  :turno #'agente-humano
  :inicio #'agente-humano-ini
  :fin 'agente-humano-fin
  :interactivo t)
