;
; Funciones para poner hang y echo a off en CMU
;

(use-package :system)
(use-package :alien)
(use-package :lisp)
(use-package :unix)
(use-package :extensions)

; Thanks to Fred Gilham for ideas and code to CMUCL functions 
; echo-hang-off, restore-echo-hang and read-char-no-hang-no-echo


(defvar echo-hang t
  "Variable que indica estado echo-hang de la salida/entrada")

(defvar old-gc-verbose nil)

(defvar fd nil)

(defvar old-c-lflag nil)

(defun echo-hang-off (&optional (stream *tty*))
  (setf echo-hang nil)
  (setf old-gc-verbose *gc-verbose*)
  (setf *gc-verbose* nil)
  (setf fd (fd-stream-fd (if (lisp::two-way-stream-p stream) 
			     (two-way-stream-output-stream stream) 
			   stream)))
  (when (unix-isatty fd)
    (with-alien ((tios (struct termios)))
		(multiple-value-bind
		    (val err)
		    (unix-tcgetattr fd (alien-sap tios))
		  (when (null val) 
		    (error "Could not tcgetattr, unix error ~S." 
			   (get-unix-error-msg err))))
		(setf old-c-lflag (slot tios 'c-lflag))
		(setf (slot tios 'c-lflag) 
		      (logand (slot tios 'c-lflag)
			      (lognot (logior tty-icanon
					      tty-echo
					      tty-echoe
					      tty-echok
					      tty-echonl))))
		(setf (deref (slot tios 'c-cc) vtime) 0
		      (deref (slot tios 'c-cc) vmin) 1)
		(multiple-value-bind 
		    (val err)
		    (unix-tcsetattr fd tcsaflush (alien-sap tios))
		  (when (null val)
		    (error "Could not tcsetattr, unix error ~S."
			   (get-unix-error-msg err)))))))

(defun restore-echo-hang ()
  (setf echo-hang t)
  (when (unix-isatty fd)
    (with-alien ((tios (struct termios)))
		(multiple-value-bind
		    (val err)
		    (unix-tcgetattr fd (alien-sap tios))
		  (when (null val)
		    (error "Could not tcgetattr, unix error ~S."
			   (get-unix-error-msg err))))
		(setf (slot tios 'c-lflag) old-c-lflag)
		(multiple-value-bind
		    (val err)
		    (unix-tcsetattr fd tcsaflush (alien-sap tios))
		  (when (null val)
		    (error "Could not tcsetattr, unix error ~S."
			   (get-unix-error-msg err))))))
  (setf *gc-verbose* old-gc-verbose)
  (values))

(defun read-char-no-hang-no-echo (&optional (stream *tty*))
  (let ((input nil))
    (unwind-protect 
	(setf input (read-char-no-hang stream))
      input)))

(defun unread-char-no-hang-no-echo (char &optional (stream *tty*))
  (unread-char char stream))
