;
; Funciones para poner hang y echo a off en SBCL
;

; Thanks to Richard M. Kreuter for ideas and code to SBCL functions 
; echo-hang-off, restore-echo-hang and read-char-no-hang-no-echo


(require :sb-posix)

(defvar echo-hang t
  "Variable que indica estado echo-hang de la salida/entrada")

(defvar stream-echo-hang nil)

(defvar old-c-lflag nil)
(defvar old-cc-vtime nil)
(defvar old-cc-vmin nil)

(defun echo-hang-off (&optional (stream sb-sys:*tty*))
  (setf echo-hang nil)
  (setf stream-echo-hang stream)
  (let ((tios (sb-posix:tcgetattr stream-echo-hang)))
    (setf old-c-lflag (sb-posix:termios-lflag tios))
    (setf (sb-posix:termios-lflag tios)
	  (logand (sb-posix:termios-lflag tios)
		  (lognot (logior sb-posix:icanon
				  sb-posix:echo
				  sb-posix:echoe
				  sb-posix:echok
				  sb-posix:echonl))))
    (setf old-cc-vtime (aref (sb-posix:termios-cc tios) sb-posix:vtime)
	  old-cc-vmin (aref (sb-posix:termios-cc tios) sb-posix:vmin))
    (setf (aref (sb-posix:termios-cc tios) sb-posix:vtime) 0
	  (aref (sb-posix:termios-cc tios) sb-posix:vmin) 1)
    (sb-posix:tcsetattr stream-echo-hang sb-posix:tcsanow tios)))

(defun restore-echo-hang ()
  (setf echo-hang t)
  (let ((tios (sb-posix:tcgetattr stream-echo-hang)))
    (setf (sb-posix:termios-lflag tios) old-c-lflag)
    (setf (aref (sb-posix:termios-cc tios) sb-posix:vtime) old-cc-vtime
	  (aref (sb-posix:termios-cc tios) sb-posix:vmin) old-cc-vmin)
    (sb-posix:tcsetattr stream-echo-hang sb-posix:tcsanow tios)))


(defun read-char-no-hang-no-echo (&optional (stream sb-sys:*tty*))
  (let ((input nil))
    (unwind-protect 
	(setf input (read-char-no-hang stream))
      input)))

(defun unread-char-no-hang-no-echo (char &optional (stream sb-sys:*tty*))
  (unread-char char stream))


;;  (load "echo-hang-off-sbcl.lsp")
