;; Programa: Package con agentes jugadores generadores de
;;           plantas carnivoras de PlanTron
;;
;; Asignatura: Introducción a la Inteligencia Artificial
;;             ETSII - ULL
;; Práctica: (sistemas de producción)
;; Profesor: Patricio García (patricio AT etsii.ull.es)
;; Fecha:    25/04/2008
;; Comentarios:
;;
;;           Los agentes definidos utilizan el entorno de juego
;;           PlanTron, definido principalmente por los packages
;;           PT, PT-BASE Y ENVIO.
;;
;;           Las funciones accesibles más importantes son:
;;
;;           ....
;;
;;           Ver la documentación existente para cada una de ellas.
;;
;;           Para su utilización es aconsejable usar el cargador
;;           del entorno de juego PlanTron: "pt-loader.lisp".
;;
;; Agradecimientos:
;;           En gran parte este agente ha sido inspiradon en el 
;;           entorno y código de "GA applet" de David Eck:
;;           http://math.hws.edu/xJava/GA/
;;
;; Dependencias:
;;
;;           Hace uso del paquete que contiene la interface (ENVIO)
;;           con el control del juego (PT-BASE).
;;
;; Modificaciones:
;;
;;           - Ver 0.1 (25/04/2008): Primera versión. Generada a
;;             partir de modificación de versión 0.2.3 de 
;;             pt-agentes-genpla.lisp
;;
;;           - Ver 0.1.1 (28/04/2008): *plantar-sobre-rastro*
;;             cambiado por *prob-plantar-sobre-rastro*
;;
;;           - Ver 0.2 (09/04/2009): Definición usando defagent.
;;             Posiblidad de jugadores gendir muertos.
;;
;;           - Ver 0.2.1 (15/04/2009): Eliminar pla-car y otros
;;             cambios menores
;;
;;
;; Por hacer:
;;           - Corrección de Bugs
;;             ------------------
;;           - En ocasiones gp-cumulo pone una planta sobre rastro con
;;             *plantar-sobre-rastro* a nil 
;;           - Que se pueda llamar con varios generadores de plantas
;;             del mismo o distinto tipo (conflictos fuentes, aleat, etc.)
;;           - En los cálculos del número de plantas inicial mínimo
;;             necesario para poder llenar todo el espacio no se tuvo 
;;             en cuenta la creación de nuevas plantas
;;           - En agente 4 a medida que se ejecutan partidas se acelera
;;             la generación de plantas
;;
;;           - Mejoras en funcionalidades
;;             --------------------------
;;           - Afinar valores por defecto de variables: sin mensajes
;;             de fallos pero rápidos
;;           - Añadir más tipos de agentes generadores
;;           - Que las plantas mueran por si solas tras un tiempo de
;;             vida no fijo
;;           - Generadores más simples, que no creen nuevas plantas,
;;             solo creen plantas iniciales
;;           - Que los agentes generadores sean reentrantes,esto es
;;             se puedan llamar al mismo agente como varios jugadores
;;             con variables de comportamiento distintas y no haya 
;;             conflictos con ellas
;;           - Aparece muchas veces "No encuentro hueco ..."
;;
;;           - Mejoras en estilo de programación
;;             ---------------------------------
;;           - Que haya una variable random por cada generador o id
;;           - Implementar Fase 2 de busca-libre-cerca
;;
;;


;;
;; Definición del package
;; ----------------------

(defpackage "PT-AGENTES-GENPLC"
  (:nicknames "PT-A-GPC")
  (:use "COMMON-LISP")
  (:export "AGENTES" "VERSION" "AGENTE-GP-ALEAT" "AGENTE-GP-ALEAT-INI" "AGENTE-GP-CUMULOS" "AGENTE-GP-CUMULOS-INI" "*RATIO-PLANTAS*" "*RATIO-MAX-INTENTOS-BPL*" "*ENERGIA-PLA*" "*RENACER*" "*PROB-PLANTAR-SOBRE-RASTRO*" "*RATIO-FUENTES-PLANTAS*" "*RATIO-MAX-INTENTOS-BLC*" "*RATIO-FUENTES*" "*PROB-MOD-FUENTE*" "*VAR-SIGM-HACIA*"))


(in-package "PT-AGENTES-GENPLC")



;;
;; Constante versión
;; -----------------

(defconstant version "0.2.1"
  "Versión del package PT-AGENTES-GENPLC.")


;;
;; Variables globales
;; ------------------

(defvar *for-cab-gpc* 3
  "Fortaleza de la cabeza de los generadores de plantas")

(defvar *for-col-gpc* 3
  "Fortaleza de la cabeza de los generadores de plantas")

(defvar *ratio-plantas* 0.025
  "Ratio de número de plantas por unidad de superficie")

(defvar *ratio-max-intentos-bpl* 0.06
  "Ratio del máximo número de intentos aleatorios de buscar posición libre para crear planta")

(defvar *energia-pla* 10
  "Energía por planta. Al crear una planta el generador gana esa energía, cuando muere una planta el generador pierde esa energía, transfiriéndola al comedor")

(defvar *prob-plantar-sobre-rastro* 0.2
  "Indica la probabilidad de plantar una planta sobre el rastro de un jugador o no.")

(defvar *renacer* t
  "Indica si al morir una planta renace nuevamente en otro lado.")

(defvar num-jug-gdir nil
  "Número de jugadores generadores de direccion en la partida actual.")

(defvar dim-tab nil
  "Dimensiones del tablero")

(defvar max-intentos-bpl nil
  "Máximo número de intentos aleatorios de buscar posición libre para crear planta") 

(defvar random-state-gp (make-random-state *random-state*)
  "Estado random para el agente-gp-aleat.")


;;
;; Funciones de apoyo
;; ------------------

(defun posible-plantar-p (tipo-cont)
  "Mira si es posible plantar una planta dado el contenido de tipo TIP-CONT de una casilla."
  (or (null tipo-cont)
      (and (eq tipo-cont 'envio:jugador)
	   (< (random 1.0 random-state-gp) *prob-plantar-sobre-rastro*))))

(defun inc-pos (pos dir)
  "Incrementa la posición POS en dirección DIR."
  (case dir
    (0 (list (first pos) (1- (second pos))))
    (1 (list (1+ (first pos)) (second pos)))
    (2 (list (first pos) (1+ (second pos))))
    (3 (list (1- (first pos)) (second pos)))))

(defun dir-inc (dir dir-ini)
  "Calcula la dirección DIR relativa a DIR-INI, esto es, gira DIR-INI pasos en sentido agujas del reloj sobre DIR."
  (rem (+ dir dir-ini) 4))

(defun elige1-y-limpia (lista)  ; No usado de momento
  "Elige aleatoriamente un elemento de la lista LISTA y devuelve dicho elemento y una nueva lista igual a LISTA pero sin el elemento"
  (let ((elm (nth (random (length lista) random-state-gp)
		  lista)))
    (values elm (remove elm lista :count 1))))

(defmacro return-if-not-nil (expr)
  "Ejecuta RETURN devolviendo EXPR solo si EXPR es distinto de NIL, sino devuelve NIL"
  `(let ((res-expr ,expr))
     (if res-expr
         (return res-expr))))

(defun gen-pos-aleat-libre (&optional (max-intentos max-intentos-bpl))
  "Busca una posición aleatoria libre"
  (let (posxy)
    (dotimes (i max-intentos nil)
      (setf posxy (list (random (first dim-tab) random-state-gp)
			(random (second dim-tab) random-state-gp)))
      (if (posible-plantar-p (envio:consulta-tipo (envio:consulta-pos-tab posxy)))
	  (return posxy)))))

(defun gen-planta (id posxy)
  "Genera una nueva planta en POSXY"
  (if posxy
      (if (envio:pet-gen-planta posxy *energia-pla*)
	  t
	(envio:my-warning "Agente ~S se le denegó generar nueva planta" id))
    (envio:mensaje "No encuentro hueco para generar nueva planta")))

;;
;; Agente generador de plantas aleatorias
;; --------------------------------------

(defun agente-gpc-aleat-ini (id tipo-juego)
  "Inicializador de agente jugador PlanTron tipo genplc que genera plantas aleatoriamente dentro del tablero."
  ; Extraemos dimensiones del tablero
  (setf dim-tab (getf tipo-juego :dim-tab))
  ; Calculamos número máximo de intentos aleatorios
  (setf max-intentos-bpl (round (* *ratio-max-intentos-bpl* (first dim-tab) (second dim-tab))))
  ; Generación de plantas
  (dotimes (i (round (* *ratio-plantas* (first dim-tab) (second dim-tab))) t)
    (gen-planta id (gen-pos-aleat-libre max-intentos-bpl))))


(defun agente-gpc-aleat (id-gpla id-gdir posicion)
  "Agente jugador PlanTron tipo genpla que genera planta aleatoriamente dentro del tablero."
  (declare (ignore posicion))
  ; Transferir energia al comedor
  (envio:pet-transf-ene id-gdir *energia-pla*)
  (if *renacer*
    ; Cuando se coman una planta generar una nueva
    (gen-planta id-gpla (gen-pos-aleat-libre max-intentos-bpl))))


(defun agente-gpc-aleat-inc (id)
  "Agente jugador PlanTron tipo genpla que genera planta aleatoriamente dentro del tablero incrementándose a cada turno."
  (gen-planta id (gen-pos-aleat-libre max-intentos-bpl)))


;;
;; Variables para agente generador de plantas en cumulos
;; -----------------------------------------------------

(defvar *ratio-fuentes-plantas* 0.1
  "Ratio de número de fuentes sobre el ratio de número de plantas por unidad de superficie")

(defvar *ratio-max-intentos-blc* 1  ; FIXME disminuir con impl. fase 2 blc
  "Ratio del máximo número de intentos de buscar libre cerca para crear planta")

(defvar *prob-mod-fuente* 0.03
  "Probabilidad de modificar fuente")

(defvar num-fuentes nil
  "Número de fuentes a utilizar")

(defvar fuentes nil
 "Array para almacenar las fuentes")

(defvar max-intentos-blc nil
  "Máximo de intentos en buscar-libre-cerca")

(defvar cont-int-blc nil
  "Contador de intentos en buscar-libre-cerca")

;;
;; Funciones para agente generador de plantas en cumulos
;; -----------------------------------------------------

(defun busca-prof-libre (posxy)
  "Busca en profundidad una casilla libre cerca de POSXY, eligiendo aleatoriamente la posición en que comenzar y limitando el número de intentos"
  (setf cont-int-blc (1- cont-int-blc))
  (let ((tipo-cont (envio:consulta-tipo (envio:consulta-pos-tab posxy))))
    (cond
     ((posible-plantar-p tipo-cont) posxy)    ; Si es posible, plantar ahí
     ((minusp cont-int-blc) nil)              ; Si fin contador, fallar
     ((eq tipo-cont 'envio:obstaculo) nil)  ; Si sobre obstáculo, fallar
     ((eq tipo-cont 'envio:jugador) nil)      ; Si sobre jugador, fallar
     ;((eq tipo-cont 'envio:planta)   ; Si sobre planta
     (t                                       ; Si sobre planta o jugador:
      ; Busca en las 4 direcciones empezando en una aleatoria
      (let ((dir-ini (random 4 random-state-gp)))
	(dotimes (i 4 nil)
	  (if (minusp cont-int-blc)
	      (return nil)
	    (return-if-not-nil
	     (busca-prof-libre (inc-pos posxy (dir-inc i dir-ini)))))))))))
     ;(t nil))))                           ; Si choca con algo, fallar


(defun busca-libre-cerca (posxy &optional (max-intentos max-intentos-blc))
  "Busca una casilla libre cerca de POSXY"
  ; Dos fases:
  ; Fase 1: Se sale con búsqueda en profundidad pisando plantas y
  ; jugadores, pero no pisando obstaculos
  (setf cont-int-blc max-intentos)
  (busca-prof-libre posxy))
  ; FIXME Fase 2: Avanza en línea recta en dirección aleatoria hasta libre

(defun gen-una-fuente (id id-fuente &optional (max-intentos max-intentos-bpl))
  "Genera la fuente aleatoria ID-FUENTE para generar plantas"
  (declare (ignore id))
  (let ((posxy (gen-pos-aleat-libre max-intentos)))
    (if posxy
	(setf (aref fuentes id-fuente)
	      posxy)
      (envio:mensaje "No puedo generar nuevo cúmulo"))))

(defun gen-fuentes (id &optional (max-intentos max-intentos-bpl))
  "Genera las fuentes para generar plantas"
  (setf num-fuentes (round (* *ratio-fuentes-plantas*
			      *ratio-plantas* 
			      (first dim-tab) (second dim-tab))))
  (setf fuentes (make-array num-fuentes :initial-element nil))
  (dotimes (i num-fuentes t)
    (gen-una-fuente id i max-intentos)))

(defun gen-planta-cumulo (id
			  &optional 
			  (max-intentos-1 max-intentos-blc)
			  (max-intentos-2 max-intentos-bpl))
  "Genera planta en cumulo en base a una fuente elegida aleatoriamente"
  (let ((id-fuente (random num-fuentes random-state-gp)))
    ; Genera planta según fuente elegida
    (gen-planta id
		(busca-libre-cerca (aref fuentes 
					 id-fuente)
				   max-intentos-1))
    ; Ver si modificar fuente
    (if (< (random 1.0 random-state-gp) *prob-mod-fuente*)
	; Plantas comienzan a crecer en otro lado
	(gen-una-fuente id id-fuente max-intentos-2))))


;;
;; Agente generador de plantas en cumulos
;; --------------------------------------

(defun agente-gpc-cumulos-ini (id tipo-juego)
  "Inicializador de agente jugador PlanTron tipo genpla que genera plantas en cúmulos dentro del tablero."
  ; Extraemos dimensiones del tablero
  (setf dim-tab (getf tipo-juego :dim-tab))
  ; Calculamos número máximo de intentos
  (setf max-intentos-bpl (round (* *ratio-max-intentos-bpl* (first dim-tab) (second dim-tab))))
  (setf max-intentos-blc (round (* *ratio-max-intentos-blc* (first dim-tab) (second dim-tab))))
  ; Generación de fuentes
  (gen-fuentes id max-intentos-bpl)
  ; Generación de plantas
  (dotimes (i (round (* *ratio-plantas* (first dim-tab) (second dim-tab))) t)
    (gen-planta-cumulo id)))


(defun agente-gpc-cumulos (id-gpla id-gdir posicion)
  "Agente jugador PlanTron tipo genpla que genera planta en cúmulos dentro del tablero."
  (declare (ignore posicion))
  ; Transferir energia al comedor
  (envio:pet-transf-ene id-gdir *energia-pla*)
  (if *renacer*
    ; Cuando se coman una planta generar una nueva
    ; Seleccionar una fuente aleatoriamente y buscar cerca,
    ; plantar y ocasionalmente cambiar fuente
    (gen-planta-cumulo id-gpla)))


(defun agente-gpc-cumulos-inc (id)
  "Agente jugador PlanTron tipo genpla que genera planta en cúmulos dentro del tablero incrementándose a cada turno."
  (gen-planta-cumulo id))


;;
;; Variables para agente generador de plantas en cumulos carnivoros
;; ----------------------------------------------------------------

(defvar *var-sigm-hacia* 128.0
  "Varianza de la sigmoide que calcula la desviación en la dirección preferida al dirigirse hacia el jugador en agente generador de plantas en cúmulos carnivoros")

;;
;; Funciones para agente generador de plantas en cumulos carnivoros
;; ----------------------------------------------------------------

(defun direccion-recta (pos1 pos2)
  "Calcula la dirección  en linea recta entre POS1 y POS2, devuelve la dirección expresada con un flotante entre [0-4) según N=0, E=1, S=2 y O=3"
  (/ (- pi (atan (- (first pos2) (first pos1)) 
		 (- (second pos2) (second pos1))))
     (/ pi 2)))

(defun random-hacia-0 (dir)
  "Redondea la dirección DIR expresada con un flotante entre [0-4)"
    (round dir))

(defun random-hacia (dir)
  "A la dirección DIR expresada con un flotante entre [0-4) según N=0, E=1, S=2 y O=3 le suma un valor aleatorio [0, 4] preferenciando cercanos a 0 ó 4 y redondea"
    (round (+ dir
	      (/ 4 (1+ (exp (* (- *var-sigm-hacia*) (- (random 1.0 random-state-gp) 0.5))))))))

(defun busca-prof-libre-hacia (posxy posxyj)
  "Busca en guiado hacia POSXYJ una casilla libre cerca de POSXY, eligiendo aleatoriamente la posición en que comenzar favoreciendi cercanas a POSXYJ y limitando el número de intentos"
  (setf cont-int-blc (1- cont-int-blc))
  (let ((tipo-cont (envio:consulta-tipo (envio:consulta-pos-tab posxy))))
    (cond
     ((equal posxy posxyj) posxy)             ; Si sobre jugador, plantar ahí
     ((posible-plantar-p tipo-cont) posxy)    ; Si es posible, plantar ahí
     ((minusp cont-int-blc) nil)              ; Si fin contador, fallar
     ((eq tipo-cont 'envio:obstaculo) nil)    ; Si sobre obstáculo, fallar
     ((eq tipo-cont 'envio:jugador) nil)      ; Si sobre jugador, fallar
     ;((eq tipo-cont 'envio:planta)   ; Si sobre planta
     (t                                       ; Si sobre planta:
      ; Busca en las 4 direcciones empezando en una aleatoria
      (let ((dir-ini (random-hacia (direccion-recta posxy posxyj))))
	(dotimes (i 4 nil)
	  (if (minusp cont-int-blc)
	      (return nil)
	    (return-if-not-nil
	     (busca-prof-libre-hacia (inc-pos posxy (dir-inc i dir-ini)) posxyj)))))))))
     ;(t nil))))                           ; Si choca con algo, fallar


(defun busca-libre-cerca-hacia (posxyf posxyj &optional (max-intentos max-intentos-blc))
  "Busca una casilla libre cerca de POSXYF en dirección preferente a POSXJ"
  ; Dos fases:
  ; Fase 1: Se sale con búsqueda en profundidad pisando plantas y
  ; jugadores, pero no pisando obstaculos
  (setf cont-int-blc max-intentos)
  (busca-prof-libre-hacia posxyf posxyj))
  ; FIXME Fase 2: Avanza en línea recta en dirección aleatoria hasta libre

(defun distancia-city (pos1 pos2)
  "Calcula la distancia city block entre POS1 y POS2"
  (+ (abs (- (first pos1) (first pos2)))
     (abs (- (second pos1) (second pos2)))))

(defun fuente-cerca ()
  "Busca la fuente que esté mas cercana a un gendir, devolviendo una lista con el ídice de la fuente y la posición del gendir mas cercano o (NIL NIL) si no localiza a ningún gendir"
  (let (dismin indmin posxymin)
    (dotimes (i num-jug-gdir (list indmin posxymin))
      (let ((posxy (envio:consulta-pos-jug (1+ i))))
	(if posxy
	    (dotimes (j num-fuentes)
	      (let ((dis (distancia-city posxy (aref fuentes j))))
		(if (or (not dismin)
			(< dis dismin))
		    (setf dismin dis
			  indmin j
			  posxymin posxy)))))))))


(defun gen-planta-cumulo-hacia (id
				&optional 
				(max-intentos-1 max-intentos-blc)
				(max-intentos-2 max-intentos-bpl))
  "Genera planta en cumulo en base a una fuente elegida ceracana a un gendir"
  (let ((idf-posxyj (fuente-cerca)))
    (if (first idf-posxyj)
	; Genera planta según fuente elegida
	(gen-planta id
		    (busca-libre-cerca-hacia (aref fuentes 
						   (first idf-posxyj))
					     (second idf-posxyj)
					     max-intentos-1))
      ; Ver si modificar fuente
      (if (< (random 1.0 random-state-gp) *prob-mod-fuente*)
	  ; Plantas comienzan a crecer en otro lado
	  (gen-una-fuente id (first idf-posxyj) max-intentos-2)))))


;;
;; Agente generador de plantas en cumulos carnivoros
;; -------------------------------------------------


(defun agente-gpc-cumcar-ini (id tipo-juego)
  "Inicializador de agente jugador PlanTron tipo genpla que genera plantas en cúmulos dentro del tablero avanzando hacia el jugador generador de direcciones mas cercano incrementándose a cada turno"
  (setf num-jug-gdir (getf tipo-juego :num-jug-gdir))
  (agente-gpc-cumulos-ini id tipo-juego))

(defun agente-gpc-cumcar (id-gpla id-gdir posicion)
  "Agente jugador PlanTron tipo genpla que genera planta en cúmulos dentro del tablero  hacia el jugador generador de direcciones mas cercano"
  (declare (ignore posicion))
  ; Transferir energia al comedor
  (envio:pet-transf-ene id-gdir *energia-pla*)
  (if *renacer*
    ; Cuando se coman una planta generar una nueva
    ; Seleccionar una fuente aleatoriamente y buscar cerca,
    ; plantar y ocasionalmente cambiar fuente
    (gen-planta-cumulo-hacia id-gpla)))

(defun agente-gpc-cumcar-inc (id)
  "Agente jugador PlanTron tipo genpla que genera planta dentro del tablero en cúmulos avanzando hacia el jugador mas cercano incrementándose a cada turno."
  (gen-planta-cumulo-hacia id))


;;
;; Definición de agentes presentes en este package
;; -----------------------------------------------

(pt:defagente agente-gpc-aleat
  "Agente jugador PlanTron tipo genpla que genera planta aleatoriamente dentro del tablero"
  :es-comido #'agente-gpc-aleat
  :inicio #'agente-gpc-aleat-ini
  :for-cab *for-cab-gpc*
  :for-col *for-col-gpc*
  :tipo-ag 'envio:gen-pla)

(pt:defagente agente-gpc-cumulos
  "Agente jugador PlanTron tipo genpla que genera planta en cúmulos dentro del tablero."
  :es-comido #'agente-gpc-cumulos
  :inicio #'agente-gpc-cumulos-ini
  :for-cab *for-cab-gpc*
  :for-col *for-col-gpc*
  :tipo-ag 'envio:gen-pla)

(pt:defagente agente-gpc-aleat-inc
  "Agente jugador PlanTron tipo genpla que genera planta aleatoriamente dentro del tablero incrementándose a cada turno"
  :turno #'agente-gpc-aleat-inc
  :es-comido #'agente-gpc-aleat
  :inicio #'agente-gpc-aleat-ini
  :for-cab *for-cab-gpc*
  :for-col *for-col-gpc*
  :tipo-ag 'envio:gen-pla)

(pt:defagente agente-gpc-cumulos-inc
  "Agente jugador PlanTron tipo genpla que genera planta aleatoriamente dentro del tablero incrementándose a cada turno"
  :turno #'agente-gpc-cumulos-inc
  :es-comido #'agente-gpc-cumulos
  :inicio #'agente-gpc-cumulos-ini
  :for-cab *for-cab-gpc*
  :for-col *for-col-gpc*
  :tipo-ag 'envio:gen-pla)

(pt:defagente agente-gpc-cumcar-inc
  "Agente jugador PlanTron tipo genpla que genera planta dentro del tablero en cúmulos avanzando hacia el jugador mas cercano incrementándose a cada turno"
  :turno #'agente-gpc-cumcar-inc
  :es-comido #'agente-gpc-cumcar
  :inicio #'agente-gpc-cumcar-ini
  :for-cab *for-cab-gpc*
  :for-col *for-col-gpc*
  :tipo-ag 'envio:gen-pla)
