;; Programa: Cargador de entorno de juego PlanTron.
;;
;; Asignatura: Introducción a la Inteligencia Artificial
;;             ETSII - ULL
;; Práctica: (sistemas de producción)
;; Profesor: Patricio García (patricio AT etsii.ull.es)
;; Fecha:    28/04/2006
;; Comentarios:
;;           Permite cargar los ficheros necesarios para el entorno
;;           del juego PlanTron.
;;
;;
;; Modificaciones:
;;
;;           - Ver 0.6 (23/04/2007): Primera versión. Extraida de
;;           "pt-loader.lisp" a raiz de dividir este fichero en
;;           tres partes: "pt-loader.lisp", "plantron.lisp" y 
;;           "pt-arena.lisp".
;;
;;           - Ver 1.0 (1/05/2007): Ajustes para versión con nueva
;;             interface de comunicaciones en package "ENVIO"
;;
;;           - Ver 1.0.1 (8/05/2007): Activación pt-arena.
;;
;;           - Ver 1.0.2 (15/05/2007): Incluida carga agentes por
;;             defecto.
;;
;;           - Ver 1.0.3 (09/04/2009): Modificado version-pt
;;
;;
;; Por hacer:
;;           - Corrección de Bugs
;;             ------------------
;;           - Modificar version-pt para incluir más versiones,
;;             por ejemplo la del fichero "VERSION"
;;

;;
;; Variables version
;; ------------------

(defconstant version-pt-loader "1.0.4" "Versión de pt-loader.")

(defun version-pt-loader ()
  "Versión del pt-loader que se ha cargado."
  version-pt-loader)


;;
;; Variables directorios
;; ---------------------

(defvar path-pt
  (merge-pathnames
   (directory-namestring
    (eval (or (find-symbol "*LOAD-TRUENAME*")
	      (find-symbol "*LOAD-PATHNAME*")
	      (find-symbol "*LOAD-PATHNAME*" "SYSTEM") ; Para GCL
	      ""))))
  "Directorio que contiene los distintos ficheros necesarios para el juego Pt.")

(defconstant path-pt-sprod
  (if (directory (merge-pathnames (pathname "sprod") path-pt))
      (merge-pathnames (pathname "sprod/") path-pt)
    path-pt)
  "Directorio que contiene los packages para implementar sistemas de producción y búsquedas.")


;;
;; Caraga de paquetes shell, generic-search y a_star
;; -------------------------------------------

;(defun load-if-exist (file pathname)
;  "Carga el fichero en PATHNAME si existe."
  ; FIXME llamada a pt-arena::match-lisp-types
;  (if (pt-arena::match-lisp-types file pathname)
;      (load (merge-pathnames file pathname))))

(if (directory (merge-pathnames "*.*" path-pt-sprod))
    (progn
      (load (merge-pathnames (pathname "shell") path-pt-sprod))
      (load (merge-pathnames (pathname "generic-search") path-pt-sprod))
      (load (merge-pathnames (pathname "a_star") path-pt-sprod))))


;;
;; Caraga de paquetes pt-base y envio
;; ----------------------------------

(load (merge-pathnames (pathname "pt-base-package") path-pt))
(load (merge-pathnames (pathname "envio-package") path-pt))
(load (merge-pathnames (pathname "pt-package") path-pt))
(load (merge-pathnames (pathname "pt-arena-package") path-pt))
(load (merge-pathnames (pathname "pt-base") path-pt))
(load (merge-pathnames (pathname "envio") path-pt))
(load (merge-pathnames (pathname "pt") path-pt))
(load (merge-pathnames (pathname "pt-arena") path-pt))


;;
;; Caraga de agentes por defecto
;; -----------------------------

(pt-arena:load-ag "pt-agentes-humanos")

(pt-arena:load-ag "pt-agentes-simples")

;(pt-arena:load-ag "pt-agentes-medios")


;;
;; Función de versiones cargadas
;; -----------------------------

(defun version-pt ()
  "Imprime las versión del los distintos módulos Pt que se han cargado."
  (nconc
   `(("PT-BASE"   ,(pt-base:version))
     ("ENVIO"     ,(envio:version))
     ("PT"        ,(pt:version))
     ("PT-ARENA"  ,(pt-arena:version))
     ("pt-loader" ,(version-pt-loader)))
   (mapcar (lambda (pn)
	       (list pn
		     (eval (find-symbol "VERSION" pn))))
	    pt::lis-pkg-ag-load)))

